import os
import sys
import numpy as np
from astropy.io import fits
from astropy import wcs
from astropy.wcs import WCS
from astropy.io.fits import update
from astropy.io.fits import getdata
import glob
from fitsio import FITS

from astropy.table import QTable, Table, Column
from astropy import units as u

from astropy.coordinates import SkyCoord # High-level coordinates
from astropy.coordinates import ICRS, Galactic, FK4, FK5 # Low-level frames
from astropy.coordinates import Angle, Latitude, Longitude # Angles


import statistics
import shutil



def create_folder(folder):
    if not (os.path.exists(folder)):
        os.makedirs(folder)

def remove_file(filename):
    if(os.path.isfile(filename)==True):
        os.remove(filename)            

def do_evtool_esass(events=None,outfile=None,evlist=None,gti=None):

    eventfiles=None
    if(events):
        eventfiles="eventfiles=\'{}\'".format((" ").join(events))
    if(evlist):
        eventfiles="eventfiles=@{}".format(evlist)

    if not (eventfiles):
        print("ERROR: Event files not provided")

    gti="gti=\'{}\'".format(gti) if(gti) else ''

    cmd=["evtool",
         eventfiles,
         gti,
        "outfile={}".format(outfile),
        "image=yes",
        "flag=0x2000",
        "pattern=15"
        ]
    
    print((" ").join(cmd))
    os.system((" ").join(cmd))

def set_bit(value, bit):
    return value | (1<<bit)

def clear_bit(value, bit):
    return value & ~(1<<bit)

def do_badpix_tm6(filename):
    """
    special corrrection for TM6 camera, needed for two CalPV-UDS observations,
    which shows high background at RAWX>250 
  
    creates new file with added _badpix.fits
    """

    nraw=384
    skip_after=250
    nskip=(nraw - skip_after)
    nrows0=nraw*nskip
    
    fout=filename.replace(".fits", "_badpix.fits")
    shutil.copyfile(filename, fout)

    hdul = fits.open(fout)
    #hdul.info()
    tstart=hdul[1].header['TSTART']
    tstop=hdul[1].header['TSTOP']
    nrows=hdul[1].header['NAXIS2']


    f = FITS(fout,'rw')
    data = f['EVENTS'].read()

    #print(data.dtype)

    for row in range(nrows):
        if (data['RAWX'][row] >= skip_after):
            data['FLAG'][row]=set_bit(data['FLAG'][row],13)
    #hdr = fits[0].read_header()
    f[1].write(data)
    f[1].write_comment("Changed by Roman Krivonos")
    f[1].write_history("All RAWx>=250 marked as BAD 0x2000")

    data_badpix = f['BADPIX6'].read()
    data4 = np.zeros(nrows0,dtype=data_badpix.dtype)

    n=0
    for i in range(skip_after,nraw):
        for j in range(nraw):
            data4['RAWX'][n]=i
            data4['RAWY'][n]=j
            data4['TIMEMIN'][n]=tstart
            data4['TIMEMAX'][n]=tstop
            data4['BADFLAG'][n]=3
            data4['TYPE'][n]=3
            data4['PHAMAX'][n]=12000
            data4['PHAMIN'][n]=0        
            n=n+1
        
    f[4].append(data4)
    f.close()

def init_events(key=None, eband_selected=[0], eband_index=None,
                ra_cen=None, de_cen=None,
                emin_kev=None, emax_kev=None, subdir=None,
                do_obsmode=False,do_center=False,do_evtool=False,do_expmap=False):
    expmaps=[]
    expmap_all=[]
    infile_expmap=[]
    emin=[]
    emax=[]
    emin_ev_str=[]
    emax_ev_str=[]
    outfile_evtool=[]
    outfile_expmap=[]

    outfile_post='.fits'

    if not (subdir):
        print("subdir=?")
        return
    
    if not (os.path.exists(subdir)):
        os.makedirs(subdir)

    print("init events -- reading key {}".format(key))
    


    infile="{}.fits".format(key)
        
    if(do_obsmode==True):
        """ correct OBS_MODE in files """
            
        lockfile="{}.obsmode.lock".format(key)
        if not os.path.isfile(lockfile):
            with fits.open(infile) as hdu:
                for h in hdu:
                    extname = h.header['EXTNAME'] if "EXTNAME" in h.header else "None"
                    changeto = 'SURVEY' if ('scan' in infile) else 'POINTING'
                    if not (h.header['OBS_MODE'] == changeto):
                        print("*** {} {} --> {}".format(extname,h.header['OBS_MODE'],changeto))
                        h.header['OBS_MODE'] = changeto
                        hdu.writeto(infile,overwrite=True)
            Path(lockfile).touch()
        else:
            print("Lock file {} is found, skipping OBS_MODE correction.".format(lockfile))
            pass

    if(do_center==True):
        """ re-center original events files """
        if not (ra_cen and de_cen):
            print("Please provide center coordinates")
            sys.exit()
        cmd=["radec2xy", 
             "{}".format(infile), 
             "{}".format(ra_cen), 
             "{}".format(de_cen)]
            
        lockfile="{}.radec2xy.lock".format(key)
        if not os.path.isfile(lockfile):
            print((" ").join(cmd))
            os.system((" ").join(cmd))
            Path(lockfile).touch()
        else:
            print("Lock file {} is found, skipping radec2xy.".format(lockfile))
            pass


    outfile_evtool="{}/{}_EventList_en{}{}".format(subdir, key, eband_index, outfile_post)

    cmd=["evtool",
         "eventfiles=%s" %(infile),
         "outfile=%s" %(outfile_evtool),
         "emin=%f" %(emin_kev),
         "emax=%f" %(emax_kev),
         #"region=%s" %(region),
         "image=yes",
         "flag=0x2000",
         "pattern=15"
        ]
    # run the command
    if(do_evtool==True):
        #log = subprocess.check_call(cmd)
        print((" ").join(cmd))
        os.system((" ").join(cmd))
        """ correct OBS_MODE """
        with fits.open(outfile_evtool) as hdu:
            for h in hdu:
                extname = h.header['EXTNAME'] if "EXTNAME" in h.header else "None"
                changeto = 'SURVEY' if ('scan' in infile) else 'POINTING'
                if not (h.header['OBS_MODE'] == changeto):
                    print("*** {} {} --> {}".format(extname,h.header['OBS_MODE'],changeto))
                    h.header['OBS_MODE'] = changeto
                    hdu.writeto(outfile_evtool,overwrite=True)


    outfile_expmap="{}_ExposureMap_en{}{}".format(os.path.join(subdir,key), eband_index, outfile_post)
    if(do_expmap==True):
        cmd=["expmap",
             "inputdatasets=%s" %(outfile_evtool),
             "emin=%s" %(emin_kev),
             "emax=%s" %(emax_kev),
             "templateimage=%s" %(outfile_evtool),  
             "singlemaps=%s" %(outfile_expmap),
             "withvignetting=yes",
             "withsinglemaps=yes","withfilebadpix=yes",
             "withmergedmaps=no",
             ]
        print((" ").join(cmd))
        os.system((" ").join(cmd))

    return outfile_evtool
